package abmland.models.infrastructure;

import static abmland.common.LogExceptionHandler.handleException;
import static abmland.common.ScheduleProperties.AGENT_INIT_WATCH_TRIGGER_P;
import static abmland.common.ScheduleProperties.AGENT_STEP_WATCH_TRIGGER_P;
import static abmland.common.ScheduleProperties.NO_TEST_SCHEDULE;
import static abmland.common.ScheduleProperties.DECISION_MODE_IS_CLEVER_COND;
import static org.apache.log4j.Logger.getLogger;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.engine.watcher.Watch;
import repast.simphony.engine.watcher.WatcherTriggerSchedule;
import repast.simphony.space.grid.GridPoint;
import abmland.common.RepastUtils;
import abmland.data.CLCover;
import abmland.data.Infrastructuretypes;
import abmland.mainctrl.config.ConfManager;
import abmland.models.ILU;
import abmland.models.IPlannerModel;
import abmland.models.IResidentModel;
import abmland.models.config.InfrastructureConf;
import abmland.types.IBooleanAllCells;

public class InfrastructureAgent {

	private final Logger log = getLogger(getClass().getName());
	private Env env = new Env();
	public final static double AGENT_RET_PROPOSALS_DATA_TRIGGER_P = ScheduleParameters.FIRST_PRIORITY;
	public final static double AGENT_RET_LUC_WITHOUT_PERMISSION_TRIGGER_P = ScheduleParameters.FIRST_PRIORITY;
	public final static double AGENT_SUMMARISE_DATA_TRIGGER_P = ScheduleParameters.FIRST_PRIORITY;
	public final static String AGENT_IS_PUBLICTRANSPORTPROVIDER_TRIGGER_COND = "$watcher.getName() == \"PUBLICTRANSPORTPROVIDER\"";
	public final static String WATCHEE_IS_PUBLICTRANSPORTPROVIDER_TRIGGER_COND = "$watchee.getName() == \"PUBLICTRANSPORTPROVIDER\"";
	private final static String INFRASTRUCTURE_WATCHEE_NAME = "abmland.models.infrastructure.InfrastructureModel";
	private final static String INFRASTRUCTUREAGENT_WATCHEE_NAME = "abmland.models.infrastructure.InfrastructureAgent";
	private final static String INIT = "initProvidersAllowed";

	final class Env {
		InfrastructureConf config;
		Map<GridPoint, Boolean> PARAM_INITIAL_EXIST_MAP;
		Map<GridPoint, Boolean> existMap;
		String HELLO = "agent-type %s (object-id: %s)";
		String CHELLO = "agent-type %s (object-id: %s) constructed";
		Infrastructuretypes iftype;
		Class<?> provider;
		InfrastructureModel model;
		abmland.models.infrastructure.InfrastructureModel.Env modelenv;
		ILU landusesrv;
		IResidentModel residentsrv;
		IPlannerModel plannersrv;
		IBooleanAllCells availabilityOfPublicTransport_map;
		Map<GridPoint, CLCover> landUseChangesWithoutPermission;
		PublicTransportProvider publicTransportProvider = new PublicTransportProvider();

		IBooleanAllCells availabilityOfPublicTransport_tempmap;

		int[] caseStudyDimensions;
	}

	public final Infrastructuretypes getIfType() {
		try {
			assertTrue(env.iftype.getDeclaringClass() == Infrastructuretypes.class);
		} catch (Exception t) {
			handleException(getClass(), t);
		}
		return env.iftype;
	}

	/**
	 * Export for the Simphony-Framework (see
	 * http://www.cscs.umich.edu/research/Repast_Simphony_Documentation.pdf,
	 * "Logging Framework")
	 *
	 * @return this agent's iftype (infrastructure type) as string acting as the
	 *         name to show on the gui and elsewhere
	 */
	public final String getName() {
		return getIfType().name();
	}

	private InfrastructureModel getModel() {

		InfrastructureModel model = null;

		Iterator<InfrastructureModel> agentIterable = (Iterator<InfrastructureModel>) RepastUtils
				.findObjects(InfrastructureModel.class).iterator();

		while (agentIterable.hasNext()) {
			model = (InfrastructureModel) agentIterable.next();
		}
		return model;
	}

	public InfrastructureAgent(Infrastructuretypes iftype) {
		env.iftype = iftype;

		log.trace(String
				.format(env.CHELLO, getIfType().name(), this.hashCode()));

		try {
			ConfManager.INFRASTRUCTURECONF.readPureConfig();
			env.config = ConfManager.INFRASTRUCTURECONF.getConfig();
		} catch (Throwable t) {
			handleException(getClass(), t);
		}
	}

	private boolean initProvidersAllowed = false;

	/**
	 * Called once, when models scheduled init method reached tail, called for
	 * all of the infrastructure agent types to have shared behavior for them.
	 */
	@Watch(//
	watcheeClassName = INFRASTRUCTURE_WATCHEE_NAME, //
	watcheeFieldNames = "initAgentAllowed", //
	whenToTrigger = WatcherTriggerSchedule.IMMEDIATE, //
	scheduleTriggerDelta = 1, //
	scheduleTriggerPriority = AGENT_INIT_WATCH_TRIGGER_P)
	public void init() {
		log.info(String.format(env.HELLO, getIfType(), this.hashCode()));

		env.model = getModel();
		try {
			assertNotNull(env.model);
		} catch (Throwable e) {
			handleException(getClass(), e);
			System.exit(0);
		}
		initProvidersAllowed = !initProvidersAllowed;  
		
		return;
	}

	/**
	 * Called (implicitly) once, when models scheduled init method reached tail,
	 * differentiates between the given infrastructure agent types to have
	 * different behavior in each of them. This one should be called only for
	 * the PUBLICTRANSPORTPROVIDER type.
	 */
	@Watch(//
	watcheeClassName = INFRASTRUCTUREAGENT_WATCHEE_NAME, //
	watcheeFieldNames = INIT, //
	whenToTrigger = WatcherTriggerSchedule.IMMEDIATE, //
	scheduleTriggerDelta = 1, //
	triggerCondition = AGENT_IS_PUBLICTRANSPORTPROVIDER_TRIGGER_COND + " && "
			+ WATCHEE_IS_PUBLICTRANSPORTPROVIDER_TRIGGER_COND, //
	scheduleTriggerPriority = AGENT_INIT_WATCH_TRIGGER_P)
	public void initPUBLICTRANSPORTPROVIDER() {
		log.info(String.format(env.HELLO, getIfType(), this.hashCode()));
		try {
			assertNotNull(env.model);
		} catch (Throwable e) {
			handleException(getClass(), e);
			System.exit(0);
		}
		env.publicTransportProvider = new PublicTransportProvider();
		env.publicTransportProvider.init(env.model.getParameters(getIfType()));
		env.provider = env.publicTransportProvider.getClass();

		env.availabilityOfPublicTransport_map = env.publicTransportProvider
				.retrieveAvailabilityOfPublicTransportByAgent();
		return;
	}
	
	public void stepPUBLICTRANSPORTPROVIDER() {
		log.info(String.format(env.HELLO, getIfType(), this.hashCode()));

		env.modelenv = getModel().getEnv();

		env.publicTransportProvider.decideCLEVER(env.modelenv.importedLandUse,
				env.modelenv.importedTotalHouseholdNumber);
		return;
	}


	/**
	 * Returns number of public transport stops for GUI.
	 */
	public int retrieveNumOf(Infrastructuretypes iftype) {
		switch (iftype) {
		case PUBLICTRANSPORTPROVIDER:
			return env.publicTransportProvider.numOf();
		default:
			break;
		}
		try {
			throw new Throwable(
					"No supported infrastructure type for this method "
							+ iftype);
		} catch (Throwable t) {
			handleException(getClass(), t);
		}
		return -1;
	}

	/**
	 * @param iftype
	 * @return a IBooleanAllCells map or null if infrastructure type given isn't
	 *         supported or no infrastructure type
	 */
	public IBooleanAllCells retrieveAvailabilityMap(Infrastructuretypes iftype) {
		switch (iftype) {
		case PUBLICTRANSPORTPROVIDER:
			return env.availabilityOfPublicTransport_map;
		default:
			break;
		}
		try {
			throw new Throwable(
					"No supported infrastructure type for this method "
							+ iftype);
		} catch (Throwable t) {
			handleException(getClass(), t);
		}
		return null;
	}

	/**
	 * Export for the Simphony-Framework (see
	 * http://www.cscs.umich.edu/research/Repast_Simphony_Documentation.pdf,
	 * "Logging Framework")
	 *
	 * @return the computed overall number of patches containing this agent's
	 *         iftype (infrastructure type)
	 */
	public int numOfPatchesThisIfTypeIsOn() {
		return retrieveNumOf(getIfType());
	}

	public int numOfPatchesThisIfTypeIsOn(Infrastructuretypes infrastructuretype) {

		return retrieveNumOf(infrastructuretype);
	}

	public Env getEnv() {
		return env;
	}
}
