package abmland.models.infrastructure;

import static abmland.common.LogExceptionHandler.handleException;
import static abmland.common.RepastUtils.findParent;
import static abmland.common.RepastUtils.getModelController;
import static abmland.common.ScheduleProperties.MODEL_COMPUTE_SCHED_P;
import static abmland.common.ScheduleProperties.MODEL_GETDATA_SCHED_P;
import static abmland.common.ScheduleProperties.MODEL_INIT_SCHED_P;
//import static abmland.mainctrl.ParametersController.getCaseStudy;
import static abmland.mainctrl.ParametersController.getDecisionMode;
//import static abmland.mainctrl.ParametersController.getScenario;
//import static abmland.mainctrl.ParametersController.getStartYear;
import static junit.framework.Assert.assertTrue;
import static org.apache.log4j.Logger.getLogger;
import static org.junit.Assert.assertNotNull;

import java.util.EnumMap;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.apache.log4j.Logger;

import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.engine.watcher.WatcheeInstrumentor;
import repast.simphony.space.grid.GridPoint;
import abmland.common.CellDumbMock;
import abmland.common.RepastUtils;
import abmland.data.CLCover;
//import abmland.data.CaseStudy;
import abmland.data.DecisionMode;
import abmland.data.Householdtypes;
import abmland.data.Infrastructuretypes;
//import abmland.data.Scenario;
import abmland.mainctrl.LandUseControllerServer;
import abmland.mainctrl.config.ConfManager;
import abmland.models.IInfrastructureModel;
import abmland.models.ILU;
import abmland.models.IPlannerModel;
import abmland.models.IResidentModel;
import abmland.models.config.InfrastructuretypeParameters;
import abmland.models.config.InfrastructureConf;
import abmland.models.default_impl.DefaultInfrastructureModel2;
import abmland.models.serve.InfrastructureModelServer2;
import abmland.models.serve.PlannerModelServer2;
import abmland.models.serve.ResidentModelServer2;
import abmland.types.IBooleanAllCells;
import abmland.types.ICLCoverAllCells;
import abmland.types.IHouseholdNumber;
import abmland.types.IPressure;
import abmland.types.IProposalsSet;
import abmland.types.impl.CLCoverSomeCells;
import abmland.types.impl.ProposalsSet;

public class InfrastructureModel extends DefaultInfrastructureModel2
/** Model2RepastSchedule , IInfrastructureModel */
{
	private final Logger log = getLogger(getClass().getName());
	private Env env = new Env();

	private boolean initAgentAllowed;
	private boolean summariseWithinAgentTrigger;

	/**
	 * To access from out-side package a field inside Env over
	 * getEnv().envfield, Env must be public, e. g. for @Watch condition
	 */
	/** public */
	final class Env {
		InfrastructureConf config;
		private ILU luserver;
		private IResidentModel residentserver;
		private IInfrastructureModel modelsrv;
		private InfrastructureAgent agentsrv;
		Set<InfrastructureAgent> agents = new HashSet<InfrastructureAgent>();
		Map<Infrastructuretypes, InfrastructureAgent> agents2 = new EnumMap<Infrastructuretypes, InfrastructureAgent>(
				Infrastructuretypes.class);

		ICLCoverAllCells importedLandUse;
		Map<GridPoint, Map<Householdtypes, IHouseholdNumber>> importedTotalHouseholdNumber;

//		/**
//		 * Variables that are defined by user input.
//		 */
		DecisionMode decisionMode;
	}

	public Env getEnv() {
		return env;
	}
	
	/**
	 * Invariant: The infrastructure type information is added in the
	 * agentcontext already
	 */
	private void addAgents() {
		for (InfrastructureAgent agent : findParent(InfrastructureAgent.class)
				.toArray(new InfrastructureAgent[0])) {
			env.agents.add(agent);
			env.agents2.put(agent.getIfType(), agent);
		}
		return;
	}

	public InfrastructureModel() {
		log.trace(null);
		env.modelsrv = new InfrastructureModelServer2(this);

		ConfManager.INFRASTRUCTURECONF.readPureConfig(); 
		env.config = ConfManager.INFRASTRUCTURECONF.getConfig();
		
	}

	/**
	 * This method is implemented for the interface.
	 * It only delivers random values in the toy model.
	 */
	public IBooleanAllCells getAvailabilityOfMainRoads() {
		return CellDumbMock.ABooleanAllCells.requestBooleanAllCells("mainroads");
	} 

	/**
	 * Ask the (over agent context => agent type map =>) agent for that data,
	 * which its provider computed earlier.
	 */
	public IBooleanAllCells getAvailabilityOfPublicTransport() {
		
		return env.agents2.get(Infrastructuretypes.PUBLICTRANSPORTPROVIDER)
				.retrieveAvailabilityMap(Infrastructuretypes.PUBLICTRANSPORTPROVIDER);
	}

	/**
	 * This method is implemented for the interface.
	 * It only delivers random values in the toy model.
	 */
	public IBooleanAllCells getSchoolSupply() {
		//
		return CellDumbMock.ABooleanAllCells.requestBooleanAllCells("schools");
	} 


	/**
	 * This method is implemented for the interface.
	 * It only delivers 0 values in the toy model.
	 */
	public Map<GridPoint, CLCover> getLandUseChanges() {

		return new CLCoverSomeCells(CLCover.NO_DATA,
				new GridPoint(1, 1));  
	}

	/**
	 * This method is implemented for the interface.
	 * It only delivers random values in the toy model.
	 */
	public Set<Map<GridPoint, CLCover>> getProposalsNeedingPermission() {
		return new ProposalsSet(new HashSet<Map<GridPoint, CLCover>>());
	}

	public InfrastructuretypeParameters getParameters(Infrastructuretypes iftype) {
		return (InfrastructuretypeParameters) env.config.infrastructuretypes.get(iftype);
	}


	@ScheduledMethod(//
	start = 1, //
	priority = MODEL_INIT_SCHED_P //
	)
	public void init() {
		log.info(null);

		/**
		 * Make known the external models to communicate with
		 */
		env.luserver = new LandUseControllerServer(
				getModelController(ILU.class));
		env.residentserver = new ResidentModelServer2(
				getModelController(IResidentModel.class));

		/**
		 * Get values for user input. They are constant throughout
		 * the simulation.
		 */
		env.decisionMode = getDecisionMode();

		initAgentAllowed = !initAgentAllowed;
		
		addAgents();

		env.agentsrv = getAgent();
	}

	/**
	 * call the export methods of the other agents implementations e. g.
	 * residents total number of households here
	 */
	@ScheduledMethod(//
	start = 2, //
	interval = 1.0, //
	priority = MODEL_GETDATA_SCHED_P //
	)
	public void getData() {
		log.info(null);
		env.importedTotalHouseholdNumber = env.residentserver.getTotalHouseholdNumber();
		env.importedLandUse = env.luserver.getLandUse();
	}

	/**
	 * Targeted Priority Chain descending: init:800 > getData:700 > compute:600
	 */
	@ScheduledMethod(//
	start = 2, //
	interval = 1.0, //
	priority = MODEL_COMPUTE_SCHED_P//
	)
	public void compute() {
		log.info(null);
		
		// trigger agent's decision
		env.agents2.get(Infrastructuretypes.PUBLICTRANSPORTPROVIDER).stepPUBLICTRANSPORTPROVIDER();

		// summarise export data 
		summariseExportData();
	}

	/**
	 * Asks agents to compute their decisions and summarises results into export
	 * tables.
	 */
	private void summariseExportData() {
		log.info(null);
		summariseWithinAgentTrigger = !summariseWithinAgentTrigger;
	}

	private InfrastructureAgent getAgent() {

		InfrastructureAgent agent = null;

		Iterator<InfrastructureAgent> agentIterable = (Iterator<InfrastructureAgent>) RepastUtils
				.findObjects(InfrastructureAgent.class).iterator();

		while (agentIterable.hasNext()) {
			agent = (InfrastructureAgent) agentIterable.next();
		}
		return agent;
	}
	
	public DecisionMode getDecisionMode(){
		return env.decisionMode;
	}
	
}
