package abmland.models.resident;

import static abmland.common.LogExceptionHandler.handleException;
import static abmland.common.ScheduleProperties.AGENT_INIT_WATCH_TRIGGER_P;
import static org.apache.log4j.Logger.getLogger;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.util.Iterator;
import java.util.Map;

import org.apache.log4j.Logger;

import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.engine.watcher.Watch;
import repast.simphony.engine.watcher.WatcherTriggerSchedule;
import repast.simphony.space.grid.GridPoint;
import abmland.common.RepastUtils;
import abmland.data.Householdtypes;
import abmland.mainctrl.config.ConfManager;
import abmland.models.ILU;
import abmland.models.config.ResidentConf;
import abmland.types.IHouseholdNumber;

public class ResidentAgent {

	private final Logger log = getLogger(getClass().getName());
	private Env env = new Env();
	public final static double AGENT_RET_PROPOSALS_DATA_TRIGGER_P = ScheduleParameters.FIRST_PRIORITY;
	public final static double AGENT_RET_LUC_WITHOUT_PERMISSION_TRIGGER_P = ScheduleParameters.FIRST_PRIORITY;
	public final static double AGENT_SUMMARISE_DATA_TRIGGER_P = ScheduleParameters.FIRST_PRIORITY;
	public final static String AGENT_IS_ELDERLYCOHABITATIONS_TRIGGER_COND = "$watcher.getName() == \"ELDERLYCOHABITATIONS\"";
	public final static String WATCHEE_IS_ELDERLYCOHABITATIONS_TRIGGER_COND = "$watchee.getName() == \"ELDERLYCOHABITATIONS\"";
	private final static String RESIDENT_WATCHEE_NAME = "abmland.models.resident.ResidentModel";
	private final static String RESIDENTAGENT_WATCHEE_NAME = "abmland.models.resident.ResidentAgent";
	private final static String INIT = "initHouseholdsAllowed";

	final class Env {
		ResidentConf config;
		Map<GridPoint, Boolean> PARAM_INITIAL_EXIST_MAP;
		Map<GridPoint, Boolean> existMap;
		String HELLO = "agent-type %s (object-id: %s)";
		String CHELLO = "agent-type %s (object-id: %s) constructed";
		Householdtypes hhtype;
		Class<?> provider;
		ResidentModel model;
		abmland.models.resident.ResidentModel.Env modelenv;

		ILU landusesrv;
		Map<GridPoint, Map<Householdtypes, IHouseholdNumber>> hhNumberOfTypeAllCells_map;
		ResidentAgentA elderlyCohabs = new ResidentAgentA();
		int[] caseStudyDimensions;
	}

	public final Householdtypes getHHType() {
		try {
			assertTrue(env.hhtype.getDeclaringClass() == Householdtypes.class);
		} catch (Exception t) {
			handleException(getClass(), t);
		}
		return env.hhtype;
	}

	/**
	 * Export for the Simphony-Framework (see
	 * http://www.cscs.umich.edu/research/Repast_Simphony_Documentation.pdf,
	 * "Logging Framework")
	 *
	 * @return this agent's hhtype (household type) as string acting as the
	 *         name to show on the gui and elsewhere
	 */
	public final String getName() {
		return getHHType().name();
	}

	private ResidentModel getModel() {

		ResidentModel model = null;

		Iterator<ResidentModel> agentIterable = (Iterator<ResidentModel>) RepastUtils
				.findObjects(ResidentModel.class).iterator();

		while (agentIterable.hasNext()) {
			model = (ResidentModel) agentIterable.next();
		}
		return model;
	}

	public ResidentAgent(Householdtypes hhtype) {
		env.hhtype = hhtype;

		log.trace(String
				.format(env.CHELLO, getHHType().name(), this.hashCode()));

		try {
			ConfManager.RESIDENTCONF.readPureConfig();
			env.config = ConfManager.RESIDENTCONF.getConfig();
		} catch (Throwable t) {
			handleException(getClass(), t);
		}
	}

	private boolean initHouseholdsAllowed = false;

	/**
	 * Called once, when models scheduled init method reached tail, called for
	 * all of the household agent types to have shared behavior for them.
	 */
	@Watch(//
	watcheeClassName = RESIDENT_WATCHEE_NAME, //
	watcheeFieldNames = "initAgentAllowed", //
	whenToTrigger = WatcherTriggerSchedule.IMMEDIATE, //
	scheduleTriggerDelta = 1, //
	scheduleTriggerPriority = AGENT_INIT_WATCH_TRIGGER_P)
	public void init() {
		log.info(String.format(env.HELLO, getHHType(), this.hashCode()));

		env.model = getModel();
		try {
			assertNotNull(env.model);
		} catch (Throwable e) {
			handleException(getClass(), e);
			System.exit(0);
		}
		initHouseholdsAllowed = !initHouseholdsAllowed;  
		
		return;
	}

	/**
	 * Called (implicitly) once, when models scheduled init method reached tail,
	 * differentiates between the given household agent types to have
	 * different behavior in each of them. This one should be called only for
	 * the PUBLICTRANSPORTPROVIDER type.
	 */
	@Watch(//
	watcheeClassName = RESIDENTAGENT_WATCHEE_NAME, //
	watcheeFieldNames = INIT, //
	whenToTrigger = WatcherTriggerSchedule.IMMEDIATE, //
	scheduleTriggerDelta = 1, //
	triggerCondition = AGENT_IS_ELDERLYCOHABITATIONS_TRIGGER_COND + " && "
			+ WATCHEE_IS_ELDERLYCOHABITATIONS_TRIGGER_COND, //
	scheduleTriggerPriority = AGENT_INIT_WATCH_TRIGGER_P)
	public void initELDERLYCOHABITATIONS() {
		log.info(String.format(env.HELLO, getHHType(), this.hashCode()));
		try {
			assertNotNull(env.model);
		} catch (Throwable e) {
			handleException(getClass(), e);
			System.exit(0);
		}
		env.elderlyCohabs = new ResidentAgentA();
		env.elderlyCohabs.init(env.model.getParameters(getHHType()));
		env.provider = env.elderlyCohabs.getClass();

		// initialise household number
		env.hhNumberOfTypeAllCells_map = env.elderlyCohabs.retrieveHouseholdNumberOfType();
		
//		log.error("ResidentAgent.java.initPublic...size() " + env.hhNumberOfTypeAllCells_map.size());
//		log.error("ResidentAgent.java.initPublic...cell 5, 5 " + env.hhNumberOfTypeAllCells_map.get(new GridPoint(5, 5)).get(Householdtypes.ELDERLYCOHABITATIONS).value());

		return;
	}
	
	public void stepResidentTypeA() {
		log.info(String.format(env.HELLO, getHHType(), this.hashCode()));

		env.modelenv = getModel().getEnv();

		env.elderlyCohabs.decideCLEVER();
		
		env.hhNumberOfTypeAllCells_map = env.elderlyCohabs.retrieveHouseholdNumberOfType();

		return;
	}

	/**
	 * Returns number of households for GUI.
	 */
	public int retrieveNumOf(Householdtypes hhtype) {
		switch (hhtype) {
		case ELDERLYCOHABITATIONS:
			return env.elderlyCohabs.numOf();
		default:
			break;
		}
		try {
			throw new Throwable(
					"No supported household type for this method "
							+ hhtype);
		} catch (Throwable t) {
			handleException(getClass(), t);
		}
		return -1;
	}



	/**
	 * @param hhtype
	 * @return a IBooleanAllCells map or null if household type given isn't
	 *         supported or no household type
	 */
	public Map<GridPoint, Map<Householdtypes, IHouseholdNumber>> retrieveHouseholdNumberOfAgent(Householdtypes hhtype) {

		return env.hhNumberOfTypeAllCells_map;
	}

	public Env getEnv() {
		return env;
	}
}
