package abmland.models.resident;

import static org.apache.log4j.Logger.getLogger;

import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

import repast.simphony.random.RandomHelper;
import repast.simphony.space.grid.GridPoint;
import abmland.data.Householdtypes;
import abmland.models.config.HouseholdtypeParameters;
import abmland.types.IHouseholdNumber;
import abmland.types.impl.HouseholdNumber;

import com.moseph.gis.raster.Raster;

public class ResidentAgentA {

	private Map<GridPoint, Integer> PARAM_AVAILABILITY_MAP;
	private Map<GridPoint, Map<Householdtypes, IHouseholdNumber>> hhNumberPerCell;

	private final Logger log = getLogger(getClass().getName());

	private HouseholdtypeParameters params;
	private Raster raster;
	
	private int totalHouseholdNumber = 0;

	/**
	 * Is called once per agent type by ResidentAgent
	 */
	protected final void init(HouseholdtypeParameters params) {
		log.trace(null);

		this.params = params;
		this.raster = params.getRaster();
//		log.error("ResidentAgentA() raster.getCellsize() = " + raster.getCellsize());

		PARAM_AVAILABILITY_MAP = new HashMap<GridPoint, Integer>();
		hhNumberPerCell = new HashMap<GridPoint, Map<Householdtypes,IHouseholdNumber>>();
		
		for (int x = 0; x <(int) raster.getRows(); x++) {
			for (int y = 0; y < (int) raster.getCols(); y++) {
				int value;
				value = (int) raster.getValue(x, y);

				PARAM_AVAILABILITY_MAP.put(new GridPoint(x, y), value );
			
				Map <Householdtypes, IHouseholdNumber> tempMap = new HashMap<Householdtypes, IHouseholdNumber>();
				tempMap.put(Householdtypes.ELDERLYCOHABITATIONS, new HouseholdNumber(value));
				hhNumberPerCell.put(new GridPoint(x,y), tempMap);
				
				totalHouseholdNumber += PARAM_AVAILABILITY_MAP.get(new GridPoint(x, y));			

			}
		}
		log.trace("totalHouseholdNumber after init() = " + totalHouseholdNumber);
	}



	/**
	 * CLEVER mode
	 * @param importedLandUse
	 * @param importedTotalPopulationInHH
	 */
	public final void decideCLEVER() {
		
		totalHouseholdNumber = 0;
		
		for (int x = 0; x <(int) raster.getRows(); x++) {
			for (int y = 0; y < (int) raster.getCols(); y++) {
				
				if (RandomHelper.nextIntFromTo(1, 10) > 5) 
				{
				int value = (int) RandomHelper.nextDoubleFromTo(0, 10);
				PARAM_AVAILABILITY_MAP.put(new GridPoint(x, y), value);
				
				Map <Householdtypes, IHouseholdNumber> tempMap = new HashMap<Householdtypes, IHouseholdNumber>();
				tempMap.put(Householdtypes.ELDERLYCOHABITATIONS, new HouseholdNumber(value));
				hhNumberPerCell.put(new GridPoint(x,y), tempMap);
				}
				else;

				totalHouseholdNumber += PARAM_AVAILABILITY_MAP.get(new GridPoint(x, y));			
			}
		}
		log.trace("totalHouseholdNumber = after decide() " + totalHouseholdNumber);
	}

	/**
	 * 
	 * @return absolute number of households
	 * for GUI Display
	 */
	public int numOf() {
		return totalHouseholdNumber;
	}

	protected final Map<GridPoint, Map<Householdtypes, IHouseholdNumber>> retrieveHouseholdNumberOfType() {
		return hhNumberPerCell;
	}

}

