package abmland.models.resident;

import static abmland.common.RepastUtils.findParent;
import static abmland.common.ScheduleProperties.MODEL_COMPUTE_SCHED_P;
import static abmland.common.ScheduleProperties.MODEL_GETDATA_SCHED_P;
import static abmland.common.ScheduleProperties.MODEL_INIT_SCHED_P;
import static org.apache.log4j.Logger.getLogger;

import java.util.EnumMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.space.grid.GridPoint;
import abmland.common.CellDumbMock;
import abmland.common.RepastUtils;
import abmland.data.DecisionMode;
import abmland.data.Householdtypes;
import abmland.data.Issuetypes;
import abmland.mainctrl.config.ConfManager;
import abmland.models.IResidentModel;
import abmland.models.config.HouseholdtypeParameters;
import abmland.models.config.ResidentConf;
import abmland.models.default_impl.DefaultResidentModel2;
import abmland.models.serve.ResidentModelServer2;
import abmland.types.IHouseholdNumber;
import abmland.types.ISatisfaction;

public class ResidentModel extends DefaultResidentModel2
{
	private final Logger log = getLogger(getClass().getName());
	private Env env = new Env();

	private boolean initAgentAllowed;

	/**
	 * To access from out-side package a field inside Env over
	 * getEnv().envfield, Env must be public, e. g. for @Watch condition
	 */
	/** public */
	final class Env {
		ResidentConf config;
		private IResidentModel modelsrv;
		private ResidentAgent agentsrv;
		Set<ResidentAgent> agents = new HashSet<ResidentAgent>();
		Map<Householdtypes, ResidentAgent> agents2 = new EnumMap<Householdtypes, ResidentAgent>(
				Householdtypes.class);

//		/**
//		 * Variables that are defined by user input.
//		 */
		DecisionMode decisionMode;
	}

	public Env getEnv() {
		return env;
	}
	

	/**
	 * Invariant: The resident type information is added in the
	 * agentcontext already
	 */
	private void addAgents() {
		for (ResidentAgent agent : findParent(ResidentAgent.class)
				.toArray(new ResidentAgent[0])) {
			env.agents.add(agent);
			env.agents2.put(agent.getHHType(), agent);
		}
		return;
	}

	public ResidentModel() {
		log.trace(null);
		env.modelsrv = new ResidentModelServer2(this);

		ConfManager.RESIDENTCONF.readPureConfig(); 
		env.config = ConfManager.RESIDENTCONF.getConfig();
		
	}

	
	public Map<GridPoint, Map<Householdtypes, IHouseholdNumber>> getTotalHouseholdNumber(){
		return env.agents2.get(Householdtypes.ELDERLYCOHABITATIONS).
		retrieveHouseholdNumberOfAgent(Householdtypes.ELDERLYCOHABITATIONS);
	}

	/**No satisfaction is computed within Resident model */
	public Map<GridPoint, Map<Householdtypes, Map<Issuetypes, ISatisfaction>>> getSatisfaction() {
		return CellDumbMock.ASatisfactionPerIssueOfAgent.requestSatisfactionPerIssueOfResidentAllCells();
	}

	public HouseholdtypeParameters getParameters(Householdtypes hhtype) {
		return (HouseholdtypeParameters) env.config.householdtypes.get(hhtype);
		
	}


	@ScheduledMethod(//
	start = 1, //
	priority = MODEL_INIT_SCHED_P //
	)
	public void init() {
		log.info(null);

		/**
		 * Get values for variables (user input). They are constant throughout
		 * the simulation.
		 */
		env.decisionMode = getDecisionMode();
		initAgentAllowed = !initAgentAllowed;
		addAgents();
		env.agentsrv = getAgent();
		initAgentAllowed = !initAgentAllowed;
	}

	/**
	 *Toy resident model does not import anything in getData()
	 */
	@ScheduledMethod(//
	start = 2, //
	interval = 1.0, //
	priority = MODEL_GETDATA_SCHED_P //
	)
	public void getData() {
		log.info(null);
	}

	/**
	 * Targeted Priority Chain descending: init:800 > getData:700 > compute:600
	 */
	@ScheduledMethod(//
	start = 2, //
	interval = 1.0, //
	priority = MODEL_COMPUTE_SCHED_P//
	)
	public void compute() {
		log.info(null);
		
		// trigger agent's decision
		env.agents2.get(Householdtypes.ELDERLYCOHABITATIONS).stepResidentTypeA();

//		 summarise export data 
		summariseExportData();
	}

	/**
	 * Asks agents to compute their decisions and summarises results into export
	 * tables.
	 */
	private void summariseExportData() {
		log.info(null);
	}

	private ResidentAgent getAgent() {

		ResidentAgent agent = null;

		Iterator<ResidentAgent> agentIterable = (Iterator<ResidentAgent>) RepastUtils
				.findObjects(ResidentAgent.class).iterator();

		while (agentIterable.hasNext()) {
			agent = (ResidentAgent) agentIterable.next();
		}
		return agent;
	}
	
	public DecisionMode getDecisionMode(){
		return env.decisionMode;
	}
	
}
